@@ -30,8 +30,10 @@ import requests
 import json
 import os.path
 import os
+from functools import reduce
 
-from . import Ticker, utils
+
+from . import Ticker, utils, fundamentals
 from . import shared
 from datetime import datetime
 import logging
@@ -45,6 +47,152 @@ def mkdirs(str):
             os.mkdir(pp)
 
 
+def build(df,date_column):
+    df.to_csv("xa.csv")
+    df.loc[df[date_column].isna(),date_column]=df["date"]
+    df['Datetime'] = _pd.to_datetime(df[date_column])
+    df=df.sort_values(by=['Datetime'])
+    df=df.groupby(by=['Datetime']).last()
+    #df = df.set_index('Datetime')
+
+    if "commonStockSharesOutstanding" in df.columns:
+        df["shares"]=df["commonStockSharesOutstanding"]
+        df = df.drop(['commonStockSharesOutstanding'], axis=1)
+
+    if "currency_symbol" in df.columns:
+        df = df.drop(['date',date_column,'currency_symbol'], axis=1)
+    else:
+        df = df.drop(['date',date_column], axis=1)
+    df = df.astype(float)
+    df=df.fillna(method="ffill").fillna(0)
+    # for col in df.columns:
+    #     print(col)
+    return df
+
+def xxtransform_fundamentals2df(json):
+    #filing_date
+    df = _pd.DataFrame(json['Financials']['Balance_Sheet']['quarterly']).T
+
+
+    #print("Balance-Sheet:##############")
+    bal = build(df,"filing_date")
+
+    #print(df)
+
+    cf = _pd.DataFrame(json['Financials']['Cash_Flow']['quarterly']).T
+    #print("Cashflow:##############")
+
+    cf = build(cf,"filing_date")
+
+    logging.debug(f"Cashflow…\n{cf}")
+
+        
+    inc = _pd.DataFrame(json['Financials']['Income_Statement']['quarterly']).T
+    #print("IncomeStatement:##############")
+    inc = build(inc,"filing_date")
+
+    #print(inc)
+        
+    # # Earnings
+    earn = _pd.DataFrame(json['Earnings']['History']).T
+    # #reportDate
+    earn=earn.drop(["beforeAfterMarket"],axis=1)
+    earn=earn.rename(columns={"currency":"currency_symbol"})
+    #print("Earnings:##############")
+    logging.debug(f"Earning columns\n:{earn.columns}")
+    earn = build(earn,"reportDate")
+
+    #print(earn)
+
+    # print(earn.head(5))
+
+    # Merging them together
+    #print("Concat bal,cf,inc and earn")
+    dfall = reduce(
+        lambda left,right: _pd.merge(
+            left,
+            right,
+            left_index=True, 
+            right_index=True, 
+            how='outer',
+            suffixes=('', '_drop')
+        ), 
+        [bal, cf, inc, earn]
+    )
+    #print("Concat bal,cf,inc and earn DONE")
+
+    #print(df.tail(10))
+
+    trend = _pd.DataFrame(json['Earnings']['Trend']).T
+
+    allx=[]
+    if len(trend.columns) > 0:
+        trend=trend[["date","period","earningsEstimateGrowth","earningsEstimateAvg","growth"]]
+
+        years=(trend["period"] == "+1y") | (trend["period"] == "0y")
+
+        adf_years=trend[years].copy()
+        adf_years=adf_years[["date","earningsEstimateGrowth","earningsEstimateAvg"]]
+        adf_years["Datetime"]=_pd.to_datetime(adf_years["date"]) #+ relativedelta(months=-12)
+        adf_years=adf_years.sort_values(by=['Datetime'])
+        adf_years=adf_years.set_index("Datetime")
+
+        df_years=trend[years].copy()
+        df_years["Datetimex"]=_pd.to_datetime(df_years["date"]) #+ relativedelta(months=-12)
+        df_years["Datetime"]=df_years["Datetimex"] + _pd.DateOffset(months=-12)
+        df_years["earningsEstimateGrowth_+1y"]=df_years["earningsEstimateGrowth"]
+        df_years["earningsEstimateAvg_+1y"]=df_years["earningsEstimateAvg"]
+        df_years=df_years.sort_values(by=['Datetime'])
+        df_years=df_years.set_index("Datetime")
+        df_years=df_years.drop(["earningsEstimateGrowth","earningsEstimateAvg","growth","period","Datetimex"],axis=1)
+
+        allx=[df_years,adf_years,dfall]
+        for idx in [1,2,3]:
+            xdf=trend[years].copy()
+            xdf["Datetimex"]=_pd.to_datetime(xdf["date"]) 
+            xdf["Datetime"]=xdf["Datetimex"] + _pd.DateOffset(months=12*idx)
+            xdf[f"earningsEstimateGrowth{idx}"]=xdf["earningsEstimateGrowth"]
+            xdf[f"earningsEstimateAvg{idx}"]=xdf["earningsEstimateAvg"]
+            xdf=xdf.sort_values(by=['Datetime'])
+            xdf=xdf.set_index("Datetime")
+            xdf=xdf.drop(["earningsEstimateGrowth","earningsEstimateAvg","growth","period","Datetimex"],axis=1)
+            allx.append(xdf)
+    else:
+        dfall["earningsEstimateAvg_+1y"]=0
+        dfall["earningsEstimateGrowth_+1y"]=0
+        dfall["earningsEstimateAvg"]=0
+        dfall["earningsEstimateAvg1"]=0
+        dfall["earningsEstimateAvg2"]=0
+        dfall["earningsEstimateAvg3"]=0
+        dfall["earningsEstimateGrowth1"]=0
+        dfall["earningsEstimateGrowth2"]=0
+        dfall["earningsEstimateGrowth3"]=0
+        allx=[dfall]
+
+    #print("CONCAT MAIN")
+    dftx = reduce(
+        lambda left,right: _pd.merge(
+            left,
+            right,
+            left_index=True, 
+            right_index=True, 
+            how='outer',
+            suffixes=('', '_drop')
+        ), 
+        allx
+    )
+    #print("CONCAT MAIN DONE!")
+
+    dup_cols = [i for i in dftx.columns if "date" in i or "Date" in i or "_drop" in i]
+        
+    dftx = dftx.drop(dup_cols, axis=1)
+
+    dftx=dftx.fillna(method="ffill")
+    dftx=dftx.fillna(0)
+
+    #logging.info(dftx)
+    return dftx.copy()
+
 def fillup(master,df,fix=True):
     idx=_pd.DataFrame(index=master.index,columns={"__IDX"})
     dx = _pd.concat([idx,df],axis=1, join="outer").drop(columns=["__IDX"]).fillna(method='ffill')
@@ -53,7 +201,7 @@ def fillup(master,df,fix=True):
         dx.columns=_pd.MultiIndex.from_tuples(dx.columns)
     return dx
 
-def create_levermann(dx,df,gspc,start):
+def create_levermann(dx,df,gspc,start,financial_flag=False):
     """
         dx=data
         df=fundamental
@@ -126,14 +274,25 @@ def create_levermann(dx,df,gspc,start):
 
     #2. EBIT Marge LJ, ebit_margin
     dxx2["l2_ebit_margin"]=0
-    dxx2.loc[dxx2["ebit_margin"]>0.12,"l2_ebit_margin"]=1
-    dxx2.loc[dxx2["ebit_margin"]<0.06,"l2_ebit_margin"]=-1 
+    if financial_flag == False:
+        dxx2.loc[dxx2["ebit_margin"]>0.12,"l2_ebit_margin"]=1
+        dxx2.loc[dxx2["ebit_margin"]<0.06,"l2_ebit_margin"]=-1 
 
     #3. Eigenkapitalquote LJ, equity_ratio
     dxx2["l3_equity_ratio"]=0
-    dxx2.loc[dxx2["equity_ratio"]>0.25,"l3_equity_ratio"]=1
-    dxx2.loc[dxx2["equity_ratio"]<0.15,"l3_equity_ratio"]=-1 
+    if financial_flag == False:
+        dxx2.loc[dxx2["equity_ratio"]>0.25,"l3_equity_ratio"]=1
+        dxx2.loc[dxx2["equity_ratio"]<0.15,"l3_equity_ratio"]=-1 
+    else:
+        dxx2.loc[dxx2["equity_ratio"]>0.10,"l3_equity_ratio"]=1
+        dxx2.loc[dxx2["equity_ratio"]<0.05,"l3_equity_ratio"]=-1 
 
+
+    #Stole from Piotrosky-Score
+    #f totalCashFromOperatingActivities is greater than netIncome, it gains 1 point towards the score.
+    dxx2["l_pio1_4"]=0
+    dxx2.loc[dxx2["totalCashFromOperatingActivities"] > dxx2["netIncome"],"l_pio1_4"]=1
+    
     #9 Kurs heute gg. Kurs vor 6 Monaten, 
     dxx2["l9_perf_6m"]=0
     dxx2.loc[dxx2["perf_6m"]>0.05,"l9_perf_6m"]=1
@@ -197,7 +356,25 @@ def create_levermann(dx,df,gspc,start):
     #dxx2[]
 
     #"l1_equity_ratio","l2_ebit_margin","l3_equity_ratio","l4_pe_5years","l5_pe_forward","l9_perf_6m","l10_perf_1y","l11_momentum","l12_reversal","l13_growth"
-    dxx2["levermann_score"]=dxx2["l1_equity_ratio"]+dxx2["l2_ebit_margin"]+dxx2["l3_equity_ratio"]+dxx2["l4_pe_5years"]+dxx2["l5_pe_forward"]+dxx2["l9_perf_6m"]+dxx2["l10_perf_1y"]+dxx2["l11_momentum"]+dxx2["l12_reversal"]+dxx2["l13_growth"]
+    #we do not have all for levermann
+    """
+    l_pio1_4 if totalCashFromOperatingActivities > netIncome
+    l1_equity_ratio
+    l2_ebit_margin
+    l3_equity_ratio
+    l4_pe_5years
+    l5_pe_forward
+    l9_perf_6m
+    l10_perf_1y
+    l11_momentum
+    l12_reversal
+    l13_growth
+    = 11 KPIs in total. (Levermann has 13)
+    """
+    dxx2["levermann_score"]=dxx2["l_pio1_4"]+dxx2["l1_equity_ratio"]+dxx2["l2_ebit_margin"]+dxx2["l3_equity_ratio"]+dxx2["l4_pe_5years"]+dxx2["l5_pe_forward"]+dxx2["l9_perf_6m"]+dxx2["l10_perf_1y"]+dxx2["l11_momentum"]+dxx2["l12_reversal"]+dxx2["l13_growth"]
+    
+    #dxx2.to_parquet("data/eod/levermann.parquet")
+    
     return dxx2
 
 class Resolver():
@@ -311,13 +488,26 @@ class EurConvertingResolver(Resolver):
             all_df.loc[all_df["market_cap"] >= 2000000000,"levermann_threshold_ratio"]=7
             all_df.loc[all_df["market_cap"] >= 5000000000,"levermann_threshold_ratio"]=4            
             all_df.loc[all_df["market_cap"] < 2000000000,"levermann_threshold_ratio"]=10
+
+            #we scale levermann from threshold = 0 to max = 1
+            #out scale reaches from 0 .. 11
+            #m = 1/(11-tx) c = - tx / (11-tx)
+            all_df.loc[all_df["market_cap"] >= 2000000000,"lm"]=1/(11-7)
+            all_df.loc[all_df["market_cap"] >= 2000000000,"lc"]=-7/(11-7)
+  
+  
+            all_df.loc[all_df["market_cap"] >= 5000000000,"lm"]=1/(11-4)
+            all_df.loc[all_df["market_cap"] >= 5000000000,"lc"]=-4/(11-4)            
             
-            
-            all_df["levermann_score_adjusted"]=all_df["levermann_score"] / all_df["levermann_threshold_ratio"]
+            all_df.loc[all_df["market_cap"] < 2000000000,"lm"]=1/(11-9)
+            all_df.loc[all_df["market_cap"] < 2000000000,"lc"]=-9/(11-9)
+
+            all_df["levermann_score_adjusted"]=all_df["lm"] * all_df["levermann_score"] + all_df["lc"]
+            #all_df["levermann_score_adjusted"]=all_df["levermann_score"] / all_df["levermann_threshold_ratio"]
 
 
 
-            all_df= all_df.drop(columns=["FOREX","OpenFOREX"])
+            all_df= all_df.drop(columns=["FOREX","OpenFOREX","OpenORIG","HighORIG","LowORIG","CloseORIG","Adjusted_Close_ORIG","NORM"])
 
 
             all_df.to_parquet(f"{self.root}/{ticker}.parquet")
@@ -426,8 +616,8 @@ class StockResolver(Resolver):
             except Exception as iex:
                 print(response.json())
 
-            # with open(f"{filepath}.json","w") as jfile:
-            #     jsonstring=json.dumps(response.json())
+            # en(f"{filepath}.json","w") as jfile:
+            #     jsonstring=json.dumps(reswith opponse.json())
             #     json.dump(jsonstring,jfile)
 
             jx=json.dumps(response.json())
@@ -613,7 +803,7 @@ class MissingFieldException(Exception):
     def __str__(self) -> str:
         return self.msg
 
-def process(jsonstr,filepath,jpath : str = None):
+def processOLD(jsonstr,filepath,jpath : str = None):
     if jpath is not None:
         with open(jpath) as jj:
             data=json.load(jj)
@@ -864,6 +1054,8 @@ def process(jsonstr,filepath,jpath : str = None):
 
             df=_pd.read_json(filepath)
 
+            #transform_fundamentals2df
+
             # logging.info(df)
            
             df=df.rename(columns={"date":"Date"})
@@ -872,7 +1064,7 @@ def process(jsonstr,filepath,jpath : str = None):
 
             
              
-            
+            print(df.columns)
             #Levermann Eigenkapitalrendite LJ
             df["gp"]=round(df["incomeBeforeTax"] / df["totalRevenue"],4)
             # EBIT Marge LJ
@@ -932,6 +1124,171 @@ def process(jsonstr,filepath,jpath : str = None):
     else:
         return None
 
+def process(jsonstr,filepath,jpath : str = None):
+    """transform the fundamentals into something usable for Levermann"""
+    if jpath is not None:
+        with open(jpath) as jj:
+            data=json.load(jj)
+    else:
+        data = jsonstr #json.loads(jsonstr)
+
+    df=fundamentals.load_fundamentals_json(data)
+    """
+    currency_symbol
+    eps_ttm_+1
+    netIncome
+    netIncome_ytd
+    netIncome_ttm
+    roe = Levermann 1
+    shares
+    eps_ttm
+    eps_ttm_1
+    eps_ttm_2
+    eps_ttm_3
+    ebit_marge_lj = Levermann 2
+    operatingIncome_ytd
+    totalRevenue_ytd
+    eigenkapitalrendit_ytd = Levermann 1
+    equity_ratio = Eigenkapitalquote LJ   =  Levermann 3
+    bal_qtr_totalStockholderEquity
+    bal_qtr_totalAssets
+    year
+    date
+    """
+    df=df.drop(columns=["currency_symbol","year","date"])
+    df = df.astype(float)
+
+    #print("Transformation DONE")
+    #for col in df.columns:
+    #    print(col)
+    
+    #Levermann Eigenkapitalrendite LJ
+    #OLDdf["gp"]=round(df["incomeBeforeTax"] / df["totalRevenue"],4)
+    df["gp"]=round(df["roe"],4)
+    #logging.info("L1 passed")
+    # EBIT Marge LJ
+    #OLD df["ebit_margin"]=round(df["netIncome"] / df["totalStockholderEquity"],4)
+    df["ebit_margin"]=round(df["ebit_marge_lj"],4)
+    
+    #logging.info("L2 passed")
+    # Eigenkapitalquote LJ
+    #df["equity_ratio"]=round(df["netIncome"] / df["totalStockholderEquity"],4)
+
+    #logging.info("L3 passed")
+    
+    
+    df["earningsEstimateAvg"]=df["eps_ttm"]
+    df["earningsEstimateAvg5"]=(df["eps_ttm"]+df["eps_ttm_1"]+df["eps_ttm_2"]+df["eps_ttm_3"]+df["eps_ttm_+1"]) / 5.0
+    #logging.info("Pass 9")
+    #df["estimatedGrowth"]=(df["earningsEstimateAvg_+1y"]-df["earningsEstimateAvg"])/df["earningsEstimateAvg"]
+    #logging.info("Pass 10")
+    #print("End of PROCESS")
+    return df
+
+
+def processV2(jsonstr,filepath,jpath : str = None):
+    """transform the fundamentals into something usable for Levermann"""
+    if jpath is not None:
+        with open(jpath) as jj:
+            data=json.load(jj)
+    else:
+        data = jsonstr #json.loads(jsonstr)
+
+    df=transform_fundamentals2df(data)
+    """
+    currency_symbol
+    eps_ttm_+1
+    netIncome
+    netIncome_ytd
+    netIncome_ttm
+    roe = Levermann 1
+    shares
+    eps_ttm
+    eps_ttm_1
+    eps_ttm_2
+    eps_ttm_3
+    ebit_marge_lj = Levermann 2
+    operatingIncome_ytd
+    totalRevenue_ytd
+    eigenkapitalrendit_ytd = Levermann 1
+    equity_ratio = Eigenkapitalquote LJ   =  Levermann 3
+    bal_qtr_totalStockholderEquity
+    bal_qtr_totalAssets
+    year
+    date
+    """
+    df = df.astype(float)
+
+    #print("Transformation DONE")
+    #for col in df.columns:
+    #    print(col)
+    
+    #Levermann Eigenkapitalrendite LJ
+    #OLDdf["gp"]=round(df["incomeBeforeTax"] / df["totalRevenue"],4)
+    df["gp"]=round(df["roe"],4)
+    #logging.info("L1 passed")
+    # EBIT Marge LJ
+    #OLD df["ebit_margin"]=round(df["netIncome"] / df["totalStockholderEquity"],4)
+    df["ebit_margin"]=round(df["ebit_marge_lj"],4)
+    
+    #logging.info("L2 passed")
+    # Eigenkapitalquote LJ
+    #df["equity_ratio"]=round(df["netIncome"] / df["totalStockholderEquity"],4)
+
+    #logging.info("L3 passed")
+    
+    """we use the Income and shares to calculate eps and use this as the estimate"""
+    
+    if "earningsEstimateAvg" not in df.columns.values:
+        df["earningsEstimateAvg"]=df["netIncome"] / df["shares"]
+
+
+    logging.info(df)
+
+    #logging.info("Shares passed earningsEstimateAvg")
+    df.loc[df["earningsEstimateAvg"] == 0  ,"earningsEstimateAvg"]=df["netIncome"] / df["shares"]
+    #logging.info(df)
+    df.loc[df["earningsEstimateAvg"].isna(),"earningsEstimateAvg"]=df["netIncome"] / df["shares"]
+    #logging.info("OKAY")
+    
+    #logging.info("Pass 5")
+    if "earningsEstimateAvg_+1y" not in df.columns.values:
+        df["earningsEstimateAvg_+1y"] = df["earningsEstimateAvg"]
+    else:
+        df.loc[df["earningsEstimateAvg_+1y"] == 0, "earningsEstimateAvg_+1y"]=df["earningsEstimateAvg"]
+        df.loc[df["earningsEstimateAvg_+1y"].isna(), "earningsEstimateAvg_+1y"]=df["earningsEstimateAvg"]
+    #logging.info("Pass 6")
+    if "earningsEstimateAvg1" not in df.columns.values:
+        df["earningsEstimateAvg1"]=df["earningsEstimateAvg"]
+    else:
+        df.loc[df["earningsEstimateAvg1"].isna(), "earningsEstimateAvg1"]=df["earningsEstimateAvg"]
+
+    
+    #logging.info("Pass 7")
+    if "earningsEstimateAvg2" not in df.columns.values:
+        df["earningsEstimateAvg2"]=df["earningsEstimateAvg1"]
+    else:
+        df.loc[df["earningsEstimateAvg2"].isna(), "earningsEstimateAvg2"]=df["earningsEstimateAvg1"]
+
+    if "earningsEstimateAvg3" not in df.columns.values:
+        df["earningsEstimateAvg3"]=df["earningsEstimateAvg2"]
+    else:
+        df.loc[df["earningsEstimateAvg3"].isna(), "earningsEstimateAvg3"]=df["earningsEstimateAvg2"]
+    #ogging.info("Pass 8")
+    cc=["earningsEstimateAvg1","earningsEstimateAvg2","earningsEstimateAvg3","earningsEstimateAvg","earningsEstimateAvg_+1y"]
+    #df = df.astype(float)
+    #logging.info(df[cc].dtypes)
+    df["earningsEstimateAvg5"]=(df["earningsEstimateAvg1"]+df["earningsEstimateAvg2"]+df["earningsEstimateAvg3"]+df["earningsEstimateAvg"]+df["earningsEstimateAvg_+1y"]) / 5.0
+    #logging.info("Pass 9")
+    df["estimatedGrowth"]=(df["earningsEstimateAvg_+1y"]-df["earningsEstimateAvg"])/df["earningsEstimateAvg"]
+    #logging.info("Pass 10")
+    #print("End of PROCESS")
+    return df
+
+def write_json(filepath,ticker,jsonstring):
+    with open(f"{filepath}","w") as ofile:
+        json.dump(jsonstring,ofile)
+
 class FundamentalResolver(Resolver):
     def __init__(self,root="data/eod",evaluate=False):
         super().__init__(path=f"{root}/fundamental_ext",extension="parquet")
@@ -973,6 +1330,7 @@ class FundamentalResolver(Resolver):
             okay=False
             try:
                 response=requests.get(url)
+                write_json(xpath,ticker,response.json())
                 okay=True
             except Exception as ex:
                 print(f"{ex}")
@@ -985,16 +1343,20 @@ class FundamentalResolver(Resolver):
             else:
                 jsonfile=f"{filepath}.json"
                 try:
+                    logging.debug(f"Writing {jsonfile} {filepath}")
                     dx=process(response.json(),jsonfile)
+                    logging.debug("PROCESS DONE")
+                    logging.debug(dx)
                     if dx is None or len(dx.index.values) == 0:
                         return None
                     dx.to_parquet(filepath)
                     return dx.fillna(method='ffill').interpolate(method='linear')
                 except MissingFieldException as ex:
-                    print(f"ERROR {ex}")
+                    logging.info(f"ERROR ticker:{ticker} Missing Field {ex}")
+                    
                     return None
                 except Exception as ex:
-                    print(f"ERROR other {ex}")
+                    print(f"ERROR other ticker:{ticker} {ex}")
                     return None 
 
 
@@ -1004,6 +1366,7 @@ class CoreResolver(Resolver):
         self.asset_index_map = asset_index_map
         self.stockResolver = StockResolver(root="data/eod")
         self.fundamentalResolver=FundamentalResolver(root="data/eod",evaluate=True)
+        self.infodf=_pd.read_parquet("data/eod/universe_info.parquet")
         
 
     def iresolve(self,ticker,api_token,start,end,filepath) -> _pd.DataFrame:
@@ -1023,8 +1386,19 @@ class CoreResolver(Resolver):
 
 
             index_df=index_df[index_df.index <= last_date]
+            #for Levermann we need to decide for Financial or Non-Financial
+            #infodf=_pd.read_parquet("data/eod/universe_info.parquet")
+            sector=self.infodf.loc[self.infodf["SYMBOL"] == ticker]["Sector"].values[0]
+            financial_flag=False
+            if sector == "Financials":
+                financial_flag=True
+            ddx=create_levermann(data_df,fund_df,index_df,start_dt,financial_flag)
 
-            ddx=create_levermann(data_df,fund_df,index_df,start_dt)
+            fcol=_pd.read_csv("config/fundamental_columns.csv")
+            
+            invalid_cols=fcol[fcol["valid"] == "N"]["column"].tolist()
+            logging.debug(f"DROPPING invalid cols {invalid_cols}")
+            ddx=ddx.drop(columns=invalid_cols,axis=1)
             ddx.to_parquet(filepath)
         except Exception as ex:
             print(f"Error {ex} in CORE {ticker}")
